import abc
from enum import Enum
from typing import List, Tuple, Union


class OrderDirection(Enum):
    ASC = 'ASC'
    DESC = 'DESC'


class NonesOrder(Enum):
    FIRST = 'FIRST'
    LAST = 'LAST'


OrderOptionsTuple = Union[Tuple[str], Tuple[str, OrderDirection], Tuple[str, OrderDirection, NonesOrder]]


class OrderOption:
    attribute: str
    direction: OrderDirection
    nones_order: NonesOrder

    def __init__(self, attribute: str, direction: OrderDirection, nones_order: NonesOrder = NonesOrder.FIRST):
        self.attribute = attribute
        self.direction = direction
        self.nones_order = nones_order


class OrderOptions:
    params: List[OrderOption]

    def __init__(self, *params: OrderOption):
        self.params = list(params)


class OrderOptionsBuilder:
    _params: List[OrderOption]

    def __init__(self):
        self._params = []

    def add(self, attribute: str, direction: OrderDirection = OrderDirection.ASC, nones_order: NonesOrder = NonesOrder.LAST) -> "OrderOptionsBuilder":
        self._params.append(OrderOption(attribute, direction, nones_order))
        return self

    def add_mass(self, *items: OrderOptionsTuple) -> "OrderOptionsBuilder":
        for item in items:
            self.add(*item)
        return self

    def build(self) -> OrderOptions:
        return OrderOptions(*self._params)


class OrderOptionsConverterInterface(abc.ABC):
    @abc.abstractmethod
    def convert(self, order: OrderOptions) -> OrderOptions:
        raise NotImplementedError()



import abc
import re
from enum import Enum
from typing import List, TypeVar, Generic

TResult = TypeVar('TResult')
TModel = TypeVar('TModel')

TResultFrom = TypeVar('TResultFrom')
TResultTo = TypeVar('TResultTo')

TModelFrom = TypeVar('TModelFrom')
TModelTo = TypeVar('TModelTo')


class Operator(Enum):
    E = '='
    NE = '!='
    GT = '>'
    LT = '<'
    GTE = '>='
    LTE = '<='
    LIKE = 'LIKE'
    ILIKE = 'ILIKE'
    IN = 'IN'
    NOT_IN = 'NOT_IN'


class SpecificationInterface(Generic[TModel, TResult], abc.ABC):
    @abc.abstractmethod
    def is_satisfied_by(self, model: TModel) -> TResult:
        raise NotImplementedError()


class SpecificationConverterInterface(Generic[TModelFrom, TResultFrom, TModelTo, TResultTo], abc.ABC):
    @abc.abstractmethod
    def convert(
        self,
        specification: SpecificationInterface[TModelFrom, TResultFrom],
    ) -> SpecificationInterface[TModelTo, TResultTo]:
        raise NotImplementedError()


class BaseAndSpecification(Generic[TModel, TResult], SpecificationInterface[TModel, TResult], abc.ABC):
    specifications: List[SpecificationInterface[TModel, TResult]]

    def __init__(self, *specifications: SpecificationInterface[TModel, TResult]):
        self.specifications = list(specifications)


class BaseOrSpecification(Generic[TModel, TResult], SpecificationInterface[TModel, TResult], abc.ABC):
    specifications: List[SpecificationInterface[TModel, TResult]]

    def __init__(self, *specifications: SpecificationInterface[TModel, TResult]):
        self.specifications = list(specifications)


class BaseNotSpecification(Generic[TModel, TResult], SpecificationInterface[TModel, TResult], abc.ABC):
    specification: SpecificationInterface[TModel, TResult]

    def __init__(self, specification: SpecificationInterface[TModel, TResult]):
        self.specification = specification


class BaseAttributeSpecification(Generic[TModel, TResult], SpecificationInterface[TModel, TResult], abc.ABC):
    attribute_name: str
    attribute_value: object
    operator: Operator

    def __init__(self, attribute_name: str, attribute_value: object, operator: Operator = Operator.E):
        self.attribute_name = attribute_name
        self.attribute_value = attribute_value
        self.operator = operator


class AndSpecification(Generic[TModel], BaseAndSpecification[TModel, bool]):
    def is_satisfied_by(self, model: TModel) -> bool:
        for specification in self.specifications:
            if not specification.is_satisfied_by(model):
                return False
        return True


class OrSpecification(Generic[TModel], BaseOrSpecification[TModel, bool]):
    def is_satisfied_by(self, model: TModel) -> bool:
        for specification in self.specifications:
            if specification.is_satisfied_by(model):
                return True
        return False


class NotSpecification(Generic[TModel], BaseNotSpecification[TModel, bool]):
    def is_satisfied_by(self, model: TModel) -> bool:
        return not self.specification.is_satisfied_by(model)


class AttributeSpecification(Generic[TModel], BaseAttributeSpecification[TModel, bool]):
    def is_satisfied_by(self, model: TModel) -> bool:
        model_attr = getattr(model, self.attribute_name)

        if model_attr is None and self.attribute_value is not None:
            return False

        if self.operator == Operator.E:
            if self.attribute_value is None:
                return model_attr is None
            return model_attr == self.attribute_value
        if self.operator == Operator.NE:
            if self.attribute_value is None:
                return model_attr is not None
            return model_attr != self.attribute_value
        if self.operator == Operator.GT:
            return model_attr > self.attribute_value
        if self.operator == Operator.LT:
            return model_attr < self.attribute_value
        if self.operator == Operator.GTE:
            return model_attr >= self.attribute_value
        if self.operator == Operator.LTE:
            return model_attr <= self.attribute_value
        if self.operator == Operator.LIKE:
            return self._like(str(self.attribute_value), model_attr)
        if self.operator == Operator.ILIKE:
            return self._like(str(self.attribute_value).lower(), model_attr.lower())
        if self.operator == Operator.IN:
            if isinstance(self.attribute_value, list):
                return model_attr in self.attribute_value
            raise ValueError('Attribute value must be a list')
        if self.operator == Operator.NOT_IN:
            if isinstance(self.attribute_value, list):
                return model_attr not in self.attribute_value
            raise ValueError('Attribute value must be a list')
        raise TypeError(f'Unsupported operator: {self.operator}')

    @staticmethod
    def _like(pattern: str, string: str) -> bool:
        # Replace SQL pattern wildcards (%) and (_) with regex equivalents
        pattern = pattern.replace('%', '.*').replace('_', '.')
        # Add start and end of string
        pattern = '^' + pattern + '$'
        # Check if the string matches the pattern
        return re.match(pattern, string) is not None



from typing import List, TypeVar, Generic, Optional, Iterable, Type
import abc

from abstractrepo.exceptions import ItemNotFoundException

from abstractrepo.order import OrderOptions, OrderDirection, NonesOrder, OrderOption
from abstractrepo.paging import PagingOptions
from abstractrepo.specification import SpecificationInterface

TModel = TypeVar('TModel')
TIdValueType = TypeVar('TIdValueType')
TCreateSchema = TypeVar('TCreateSchema')
TUpdateSchema = TypeVar('TUpdateSchema')


class CrudRepositoryInterface(abc.ABC, Generic[TModel, TIdValueType, TCreateSchema, TUpdateSchema]):
    @abc.abstractmethod
    def get_collection(
        self,
        filter_spec: Optional[SpecificationInterface[TModel, bool]] = None,
        order_options: Optional[OrderOptions] = None,
        paging_options: Optional[PagingOptions] = None,
    ) -> Iterable[TModel]:
        raise NotImplementedError()

    @abc.abstractmethod
    def count(self, filter_spec: Optional[SpecificationInterface[TModel, bool]] = None) -> int:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_item(self, item_id: TIdValueType) -> TModel:
        raise NotImplementedError()

    @abc.abstractmethod
    def exists(self, item_id: TIdValueType) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def create(self, form: TCreateSchema) -> TModel:
        raise NotImplementedError()

    @abc.abstractmethod
    def update(self, item_id: TIdValueType, form: TUpdateSchema) -> TModel:
        raise NotImplementedError()

    @abc.abstractmethod
    def delete(self, item_id: TIdValueType) -> TModel:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def model_class(self) -> Type[TModel]:
        raise NotImplementedError()


class ListBasedCrudRepository(
    Generic[TModel, TIdValueType, TCreateSchema, TUpdateSchema],
    CrudRepositoryInterface[TModel, TIdValueType, TCreateSchema, TUpdateSchema],
    abc.ABC,
):
    _db: List[TModel]

    def __init__(self):
        self._db = []

    def get_collection(
        self,
        filter_spec: Optional[SpecificationInterface[TModel, bool]] = None,
        order_options: Optional[OrderOptions] = None,
        paging_options: Optional[PagingOptions] = None,
    ) -> Iterable[TModel]:
        result = self._db.copy()
        result = self._apply_filter(result, filter_spec)
        result = self._apply_order(result, order_options)
        result = self._apply_paging(result, paging_options)
        return result

    def count(self, filter_spec: Optional[SpecificationInterface[TModel, bool]] = None) -> int:
        return len(self._apply_filter(self._db, filter_spec))

    def get_item(self, item_id: TIdValueType) -> TModel:
        return self._find_by_id(item_id)

    def exists(self, item_id: TIdValueType) -> bool:
        return bool(len(list(filter(lambda item: self._get_id_filter_specification(item_id).is_satisfied_by(item), self._db))))

    def create(self, form: TCreateSchema) -> TModel:
        item = self._create_model(form, self._generate_id())
        self._db.append(item)
        return item

    def update(self, item_id: TIdValueType, form: TUpdateSchema) -> TModel:
        return self._update_model(self._find_by_id(item_id), form)

    def delete(self, item_id: int) -> TModel:
        item = self._find_by_id(item_id)
        self._db = self._exclude_by_id(item_id)
        return item

    @abc.abstractmethod
    def _create_model(self, form: TCreateSchema, new_id: TIdValueType) -> TModel:
        raise NotImplementedError()

    @abc.abstractmethod
    def _update_model(self, model: TModel, form: TUpdateSchema) -> TModel:
        raise NotImplementedError()

    @abc.abstractmethod
    def _generate_id(self) -> TIdValueType:
        raise NotImplementedError()

    @abc.abstractmethod
    def _get_id_filter_specification(self, item_id: TIdValueType) -> SpecificationInterface[TModel, bool]:
        raise NotImplementedError()

    def _find_by_id(self, item_id: TIdValueType) -> TModel:
        try:
            return next(filter(lambda item: self._get_id_filter_specification(item_id).is_satisfied_by(item), self._db))
        except StopIteration:
            raise ItemNotFoundException(self.model_class, item_id)

    def _exclude_by_id(self, item_id: TIdValueType) -> TModel:
        return list(filter(lambda item: not self._get_id_filter_specification(item_id).is_satisfied_by(item), self._db))

    @staticmethod
    def _apply_filter(items: List[TModel], filter_spec: Optional[SpecificationInterface[TModel, bool]]) -> List[TModel]:
        if filter_spec is None:
            return items

        return list(filter(filter_spec.is_satisfied_by, items))

    @staticmethod
    def _apply_order(items: List[TModel], order_options: Optional[OrderOptions]) -> List[TModel]:
        if order_options is None:
            return items

        def get_none_key(order_option: OrderOption, value_is_none: bool) -> bool:
            if int(order_option.nones_order == NonesOrder.FIRST) ^ int(order_option.direction == OrderDirection.DESC):
                return not value_is_none
            else:
                return value_is_none

        def get_sort_key(order_option: OrderOption, item):
            value = getattr(item, option.attribute)
            none_key = get_none_key(order_option, value is None)
            return none_key, value

        for option in reversed(order_options.params):
            items = sorted(
                items,
                key=lambda item: get_sort_key(option, item),
                reverse=option.direction == OrderDirection.DESC,
            )

        return items

    @staticmethod
    def _apply_paging(items: List[TModel], paging_options: Optional[PagingOptions]) -> List[TModel]:
        if paging_options is None:
            return items

        return items[paging_options.offset:paging_options.offset + paging_options.limit]
